# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""CachedFile is a utility class used by AutoTransform to cache file content on read/write to
reduce excessive file reads and speed up processing. The FILE_CACHE variable stores cached
file contents, while a CachedFile object is used to interact with the cache. All writes
should go through the CachedFile object to ensure the cache is properly updated."""

from typing import Dict

FILE_CACHE: Dict[str, str] = {}


class CachedFile:

    """A wrapper that allows accessing cached file contents. Content is stored in the
    FILE_CACHE variable. Writing files outside the CachedFile write method can lead to
    invalid cache data.

    Attributes:
        path (str): The path of the file being cached.
    """

    path: str

    def __init__(self, path: str):
        """A simple constructor.

        Args:
            path (str): The path of the file being cached.
        """

        self.path = path

    @staticmethod
    def _read(path: str) -> str:
        """A simple method to read the content of a file. Used as a hook for testing
        purposes. Simply handles reading and does not interact with the cache.

        Args:
            path (str): The path to read from.

        Returns:
            str: The content of the file.
        """

        # pylint: disable=unspecified-encoding

        with open(path, "r") as file:
            content = file.read()
        return content

    def get_content(self) -> str:
        """Gets the file contents from the cache if present. If not present, it will read the
        file contents and cache it before returning.

        Returns:
            str: The contents of the file represented by this object.
        """

        if self.path not in FILE_CACHE:
            FILE_CACHE[self.path] = self._read(self.path)
        return FILE_CACHE[self.path]

    @staticmethod
    def _write(path: str, content: str) -> None:
        """A simple private method to write new content to a file. Used as a hook for testing
        purposes. Simply handles writing and does not interact with the cache.

        Args:
            path (str): The path to write the file content to.
            content (str): The content to write to the file.
        """

        # pylint: disable=unspecified-encoding

        with open(path, "w") as file:
            file.write(content)
            file.flush()

    def write_content(self, new_content: str) -> None:
        """Updates the content of a cached file, including writing the file.

        Args:
            new_content (str): The content to put in the file.
        """

        # pylint: disable=unspecified-encoding

        self._write(self.path, new_content)
        FILE_CACHE[self.path] = new_content



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A config fetcher that utilizes environment variables for storing config settings."""

import json
import os

from autotransform.config.config import Config
from autotransform.config.default import DefaultConfigFetcher
from autotransform.config.fetcher import ConfigFetcher
from autotransform.repo.base import FACTORY as repo_factory
from autotransform.runner.base import FACTORY as runner_factory


class EnvironmentConfigFetcher(ConfigFetcher):  # pylint: disable=too-few-public-methods
    """A ConfigFetcher that utilizes environment variables as configuration storage.
    Environment variable names are of the form AUTO_TRANSFORM_<SETTING> using the settings
    from the Config class. The DefaultConfigFetcher will be used for fallbacks where
    environment variables are not present. Set AUTO_TRANSFORM_CONFIG_USE_FALLBACK to "False"
    if you do not want to use a fallback.
    """

    def get_config(self) -> Config:
        """Fetch the Config.

        Returns:
            Config: The Config for AutoTransform.
        """

        local_runner_json = os.getenv("AUTO_TRANSFORM_LOCAL_RUNNER")
        if local_runner_json is not None:
            local_runner = runner_factory.get_instance(json.loads(local_runner_json))
        else:
            local_runner = None

        remote_runner_json = os.getenv("AUTO_TRANSFORM_REMOTE_RUNNER")
        if remote_runner_json is not None:
            remote_runner = runner_factory.get_instance(json.loads(remote_runner_json))
        else:
            remote_runner = None

        repo_override_json = os.getenv("AUTO_TRANSFORM_REPO_OVERRIDE")
        if repo_override_json is not None:
            repo_override = repo_factory.get_instance(json.loads(repo_override_json))
        else:
            repo_override = None

        config = Config(
            github_token=os.getenv("AUTO_TRANSFORM_GITHUB_TOKEN"),
            github_base_url=os.getenv("AUTO_TRANSFORM_GITHUB_BASE_URL"),
            jenkins_user=os.getenv("AUTO_TRANSFORM_JENKINS_USER"),
            jenkins_token=os.getenv("AUTO_TRANSFORM_JENKINS_TOKEN"),
            jenkins_base_url=os.getenv("AUTO_TRANSFORM_JENKINS_BASE_URL"),
            component_directory=os.getenv("AUTO_TRANSFORM_COMPONENT_DIRECTORY"),
            local_runner=local_runner,
            open_ai_api_key=os.getenv("AUTO_TRANSFORM_OPEN_AI_API_KEY"),
            remote_runner=remote_runner,
            repo_override=repo_override,
        )

        if os.getenv("AUTO_TRANSFORM_CONFIG_USE_FALLBACK", "true").lower() != "false":
            config = DefaultConfigFetcher().get_config().merge(config)

        return config



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Conditions are used by ConditionalStep to determine whether to take an Action.
Provides a generic way for creating Steps that take action conditionally."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The config command is used to update the config files for AutoTransform."""

import json
import os
import subprocess
from argparse import ArgumentParser, Namespace
from typing import Optional, Tuple

from autotransform.config import (
    CONFIG_FILE_NAME,
    get_cwd_config_dir,
    get_repo_config_dir,
    get_repo_config_relative_path,
)
from autotransform.config.config import Config
from autotransform.repo.base import RepoName
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.console import choose_yes_or_no, info
from autotransform.util.enums import SchemaType
from autotransform.util.manager import Manager
from autotransform.util.package import get_config_dir, get_examples_dir
from autotransform.util.scheduler import Scheduler
from autotransform.util.schema_map import SchemaMap


def add_args(parser: ArgumentParser) -> None:
    """Adds the args to a subparser that are required to initialize AutoTransform.

    Args:
        parser (ArgumentParser): The parser for the initialization.
    """

    parser.add_argument(
        "-s",
        "--simple",
        dest="simple",
        action="store_true",
        required=False,
        help="Tells the script to perform a simple set-up, selecting smart defaults.",
    )
    github_group = parser.add_mutually_exclusive_group()
    github_group.add_argument(
        "--github",
        dest="github",
        action="store_true",
        required=False,
        help="Tells the script that Github is being used.",
    )
    github_group.add_argument(
        "--no-github",
        dest="github",
        action="store_false",
        required=False,
        help="Tells the script that Github is not being used.",
    )

    jenkins_group = parser.add_mutually_exclusive_group()
    jenkins_group.add_argument(
        "--jenkins",
        dest="jenkins",
        action="store_true",
        required=False,
        help="Tells the script that Jenkins is being used.",
    )
    jenkins_group.add_argument(
        "--no-jenkins",
        dest="jenkins",
        action="store_false",
        required=False,
        help="Tells the script that Jenkins is not being used.",
    )

    parser.set_defaults(func=initialize_command_main, simple=False, github=None, jenkins=None)


# pylint: disable=too-many-arguments
def initialize_config(
    config_path: str,
    config_name: str,
    prev_config: Optional[Config] = None,
    simple: bool = False,
    use_github: Optional[bool] = None,
    use_jenkins: Optional[bool] = None,
) -> Tuple[Config, Optional[bool], Optional[bool]]:
    """Gets all of the inputs required to create a config file and write it.

    Args:
        config_path (str): The path to the config.
        config_name (str): The name of the config: user, repo, or cwd.
        prev_config (Optional[Config[, optional]): Previously input Config. Defaults to None.
        simple (bool, optional): Whether to use the simple setup. Defaults to False.
        use_github (bool, optional): Whether to use Github or not. Defaults to None.
        use_jenkins (bool, optional): Whether to use Jenkins or not. Defaults to None.

    Returns:
        Tuple[Config, Optional[bool]]: The input Config and whether it uses Github and Jenkins.
    """

    info(f"Initializing {config_name} config located at: {config_path}")

    if os.path.exists(config_path):
        reset_config = not simple and choose_yes_or_no("An existing config was found, replace it?")
        existing_config = Config.read(config_path)
        if not reset_config:
            return (existing_config, use_github, use_jenkins)
        if prev_config is not None:
            prev_config = existing_config.merge(prev_config)
        else:
            prev_config = existing_config

    config, github, jenkins = Config.from_console(
        prev_config=prev_config,
        simple=simple,
        use_github=use_github,
        use_jenkins=use_jenkins,
        user_config=config_name == "user",
    )
    config.write(config_path)
    return (config, github, jenkins)


def initialize_workflows(repo_dir: str, examples_dir: str, prev_config: Optional[Config]) -> None:
    """Set up the workflow files for using Github workflows.

    Args:
        repo_dir (str): The top level directory of the repo.
        examples_dir (str): Where example files are located.
        prev_config (Optional[Config]): Previously input Config.
    """

    component_directory = prev_config.component_directory if prev_config is not None else None
    relative_config_dir = get_repo_config_relative_path()

    workflows = [
        "autotransform.manage.yml",
        "autotransform.run.yml",
        "autotransform.schedule.yml",
        "autotransform.update.yml",
    ]
    for workflow in workflows:
        with open(f"{examples_dir}/workflows/{workflow}", "r", encoding="UTF-8") as workflow_file:
            workflow_text = workflow_file.read()

        workflow_text = workflow_text.replace("<CONFIG DIR>", relative_config_dir)

        if component_directory is not None:
            workflow_text = workflow_text.replace("<CUSTOM COMPONENTS>", component_directory)
        else:
            workflow_text = "\n".join(
                [line for line in workflow_text.split("\n") if "<CUSTOM COMPONENTS>" not in line]
            )

        workflow_file_path = f"{repo_dir}/.github/workflows/{workflow}"
        os.makedirs(os.path.dirname(workflow_file_path), exist_ok=True)
        with open(workflow_file_path, "w+", encoding="UTF-8") as workflow_file:
            workflow_file.write(workflow_text)
            workflow_file.flush()


def initialize_docker(repo_dir: str, examples_dir: str) -> None:
    """Set up the workflow files for using Github workflows.

    Args:
        repo_dir (str): The top level directory of the repo.
        examples_dir (str): Where example files are located.
    """

    relative_config_dir = get_repo_config_relative_path()

    docker_files = [
        "docker_autotransform.sh",
        "Dockerfile",
        "entrypoint.sh",
    ]
    for file in docker_files:
        with open(f"{examples_dir}/docker/{file}", "r", encoding="UTF-8") as docker_file:
            docker_file_text = docker_file.read()

        docker_file_path = f"{repo_dir}/{relative_config_dir}/docker/{file}"
        os.makedirs(os.path.dirname(docker_file_path), exist_ok=True)
        with open(docker_file_path, "w+", encoding="UTF-8") as workflow_file:
            workflow_file.write(docker_file_text)
            workflow_file.flush()


def initialize_repo(
    repo_dir: str,
    prev_config: Optional[Config],
    simple: bool = False,
    use_github: Optional[bool] = None,
    use_jenkins: Optional[bool] = None,
) -> None:
    """Set up a repo to work with AutoTransform.

    Args:
        repo_dir (str): The top level directory of the repo.
        prev_config (Optional[Config]): Previously input Config.
        simple (bool, optional): Whether to use the simple setup. Defaults to False.
        use_github (bool, optional): Whether to use Github or not. Defaults to None.
        use_jenkins (bool, optional): Whether to use Jenkins or not. Defaults to None.
    """

    examples_dir = get_examples_dir()
    repo_config_dir = get_repo_config_dir()

    github = use_github
    if github is None:
        github = choose_yes_or_no("Do you want to configure AutoTransform to use Github?")

    # Set up workflow files
    use_github_actions = github and (
        simple or choose_yes_or_no("Use Github Actions for AutoTransform?")
    )
    if use_github_actions:
        initialize_workflows(repo_dir, examples_dir, prev_config)
    elif use_jenkins:
        initialize_docker(repo_dir, examples_dir)

    # Get the Manager
    manager = Manager.init_from_console(
        repo_name=RepoName.GITHUB if github else RepoName.GIT,
        simple=simple,
    )

    # Set up the sample schema
    use_sample_schema = simple or choose_yes_or_no("Would you like to include the sample schema?")
    schema_map = SchemaMap.get()
    if use_sample_schema:
        sample_schema_file_name = "black_format.json"
        sample_schema_path = f"{examples_dir}/schemas/{sample_schema_file_name}"
        schema_map.add_schema("Black Format", SchemaType.FILE, sample_schema_file_name)
        with open(sample_schema_path, "r", encoding="UTF-8") as sample_schema_file:
            schema = AutoTransformSchema.from_data(json.loads(sample_schema_file.read()))
        schema.repo = manager.repo  # pylint: disable=protected-access

        schema_path = f"{SchemaMap.get_schema_directory()}/{sample_schema_file_name}"
        os.makedirs(os.path.dirname(schema_path), exist_ok=True)
        with open(schema_path, "w+", encoding="UTF-8") as schema_file:
            schema_file.write(json.dumps(schema.bundle(), indent=4))
            schema_file.flush()

        # Get requirements
        with open(f"{examples_dir}/requirements.txt", "r", encoding="UTF-8") as requirements_file:
            requirements = requirements_file.read()
    else:
        requirements = ""

    # Set up requirements file
    requirements_path = f"{repo_config_dir}/requirements.txt"
    os.makedirs(os.path.dirname(requirements_path), exist_ok=True)
    with open(requirements_path, "w+", encoding="UTF-8") as requirements_file:
        requirements_file.write(requirements)
        requirements_file.flush()

    # Set up schema map file
    schema_map = SchemaMap.get()
    schema_map.write()

    # Set up manage file
    manager_path = f"{repo_config_dir}/manager.json"
    manager.write(manager_path)

    # Set up schedule file
    scheduler = Scheduler.init_from_console(use_sample_schema, simple)
    scheduler_path = f"{repo_config_dir}/scheduler.json"
    scheduler.write(scheduler_path)

    if use_github_actions:
        info(
            "Please ensure there is a repository secret of AUTO_TRANSFORM_BOT_GITHUB_TOKEN "
            + "that contains a Github token for your bot account"
        )
        info(
            "This token must have permissions to push branches, "
            + "create pull requests, and trigger workflows."
        )


def initialize_command_main(args: Namespace) -> None:
    """The main method for the schedule command, handles the actual execution of scheduling runs.

    Args:
        _args (Namespace): The arguments supplied to the initialize command.
    """

    simple = args.simple
    github = args.github
    jenkins = args.jenkins

    user_config_path = f"{get_config_dir()}/{CONFIG_FILE_NAME}"
    config, github, jenkins = initialize_config(
        user_config_path, "user", None, simple=simple, use_github=github, use_jenkins=jenkins
    )

    # Set up repo
    try:
        dir_cmd = ["git", "rev-parse", "--show-toplevel"]
        repo_dir = subprocess.check_output(dir_cmd, encoding="UTF-8").replace("\\", "/").strip()
        info(f"Repo found at {repo_dir}")
        setup_repo = choose_yes_or_no("Should AutoTransform set up the repo?")
    except Exception:  # pylint: disable=broad-except
        info("No git repo to set up, run inside a git repo to set it up.")
        repo_dir = ""
        setup_repo = False

    if setup_repo:
        repo_config_path = f"{get_repo_config_dir()}/{CONFIG_FILE_NAME}"
        config, github, jenkins = initialize_config(
            repo_config_path, "repo", config, simple=simple, use_github=github, use_jenkins=jenkins
        )
        initialize_repo(repo_dir, config, simple=simple, use_github=github, use_jenkins=jenkins)

    if repo_dir == "" and choose_yes_or_no("Set up configuration for current working directory?"):
        cwd_config_path = f"{get_cwd_config_dir()}/{CONFIG_FILE_NAME}"
        initialize_config(
            cwd_config_path,
            "current working directory",
            config,
            simple=simple,
            use_github=github,
            use_jenkins=jenkins,
        )



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for regex based filters, including RegexFilter and FileContentRegexFilter."""

from __future__ import annotations

import re
from typing import ClassVar

from autotransform.filter.base import Filter, FilterName
from autotransform.item.base import Item
from autotransform.item.file import FileItem


class RegexFilter(Filter):
    """A Filter which only passes Items where the Item's key matches a provided regex pattern.
    Uses re.search rather than re.match.

    Attributes:
        pattern (str): The pattern to use when checking the Item's key.
        name (ClassVar[FilterName]): The name of the component.
    """

    pattern: str

    name: ClassVar[FilterName] = FilterName.REGEX

    def _is_valid(self, item: Item) -> bool:
        """Check whether the key contains the pattern.

        Args:
            item (Item): The Item to check.

        Returns:
            bool: Returns True if the pattern is found within the key.
        """

        return re.search(self.pattern, item.key) is not None


class RegexFileContentFilter(Filter):
    """A Filter which only passes FileItems where the file's content contains a match to the
    provided regex pattern. Uses re.search rather than re.match.

    Attributes:
        pattern (str): The pattern to use when checking the FileItem's content
        name (ClassVar[FilterName]): The name of the component.
    """

    pattern: str

    name: ClassVar[FilterName] = FilterName.REGEX_FILE_CONTENT

    def _is_valid(self, item: Item) -> bool:
        """Check whether the contents of the file contains the pattern.

        Args:
            item (Item): The Item to check.

        Returns:
            bool: Returns True if the pattern is found within the file's content.
        """

        assert isinstance(item, FileItem)
        return re.search(self.pattern, item.get_content()) is not None



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Provides utility methods for interacting with the management of outstanding Changes."""

from __future__ import annotations

import json
import os
from typing import Any, Dict, List, Optional

from autotransform.change.base import ReviewState
from autotransform.config import get_config
from autotransform.event.action import ManageActionEvent
from autotransform.event.debug import DebugEvent
from autotransform.event.handler import EventHandler
from autotransform.repo.base import FACTORY as repo_factory
from autotransform.repo.base import Repo, RepoName
from autotransform.repo.git import GitRepo
from autotransform.repo.github import GithubRepo
from autotransform.runner.local import LocalRunner
from autotransform.step.action.source import AbandonAction, MergeAction, UpdateAction
from autotransform.step.base import FACTORY as step_factory
from autotransform.step.base import Step
from autotransform.step.condition.comparison import ComparisonType
from autotransform.step.condition.state import ReviewStateCondition
from autotransform.step.condition.updated import UpdatedAgoCondition
from autotransform.step.conditional import ConditionalStep
from autotransform.util.component import ComponentModel
from autotransform.util.console import (
    choose_options_from_list,
    choose_yes_or_no,
    get_str,
    input_int,
)


class Manager(ComponentModel):
    """The information and functionality required for managing outstanding changes.

    Attributes:
        repo (Repo): The repo to get outstanding changes for.
        steps (List[Step]): The steps to take for outstanding changes.
    """

    repo: Repo
    steps: List[Step]

    def run(self, run_local: bool = False) -> None:
        """Runs the management steps.

        Args:
            run_local (bool, optional): Whether to use local runners. Defaults to False.
        """

        if run_local:
            runner = get_config().local_runner
        else:
            runner = get_config().remote_runner

        UpdateAction.set_runner(runner or LocalRunner())

        changes = self.repo.get_outstanding_changes()
        for change in changes:
            EventHandler.get().handle(DebugEvent({"message": f"Checking steps for {change!r}"}))
            for step in self.steps:
                EventHandler.get().handle(DebugEvent({"message": f"Checking step {step!r}"}))
                actions = step.get_actions(change)
                for action in actions:
                    EventHandler.get().handle(
                        ManageActionEvent({"action": action, "change": change, "step": step})
                    )
                    action.run(change)

                if actions and not step.continue_management(change):
                    EventHandler.get().handle(DebugEvent({"message": "Steps ended"}))
                    break

    def write(self, file_path: str) -> None:
        """Writes the management information to a file as JSON.

        Args:
            file_path (str): The file to write to.
        """

        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+", encoding="UTF-8") as manager_file:
            manager_file.write(json.dumps(self.bundle(), indent=4))
            manager_file.flush()

    @staticmethod
    def read(file_path: str) -> Manager:
        """Reads management info from a JSON encoded file.

        Args:
            file_path (str): The path where the JSON for the management info is located.

        Returns:
            Manager: The Manager from the file.
        """

        with open(file_path, "r", encoding="UTF-8") as manager_file:
            manager_json = manager_file.read()
        EventHandler.get().handle(
            DebugEvent({"message": f"Manager: ({file_path})\n{manager_json}"})
        )
        return Manager.from_json(manager_json)

    @staticmethod
    def from_json(manager_json: str) -> Manager:
        """Builds a Manager from JSON encoded values.

        Args:
            manager_json (str): The JSON encoded Manager.

        Returns:
            Manager: The Manager from the JSON.
        """

        return Manager.from_data(json.loads(manager_json))

    @staticmethod
    def from_data(data: Dict[str, Any]) -> Manager:
        """Produces an instance of the Manager from decoded data.

        Args:
            data (Mapping[str, Any]): The JSON decoded data.

        Returns:
            Manager: An instance of the Manager.
        """

        return Manager(
            repo=repo_factory.get_instance(data["repo"]),
            steps=[step_factory.get_instance(step) for step in data.get("steps", [])],
        )

    # pylint: disable=too-many-branches
    @staticmethod
    def init_from_console(
        repo_name: Optional[RepoName] = None,
        simple: bool = False,
    ) -> Manager:
        """Gets a Manager using console inputs.

        Args:
            repo_name (Optional[RepoName], optional): The name of the repo to use. Defaults to None.
            simple (bool, optional): Whether to use simple options for setting up the manager.
                Defaults to False.

        Returns:
            Manager: The input Manager.
        """

        if repo_name == RepoName.GITHUB:
            base_branch = get_str(
                "Enter the name of the base branch for the repo(i.e. main, master): "
            )
            github_name = get_str("Enter the fully qualified name of the github repo(owner/repo): ")
            repo: Repo = GithubRepo(base_branch=base_branch, full_github_name=github_name)
        elif repo_name == RepoName.GIT:
            base_branch = get_str(
                "Enter the name of the base branch for the repo(i.e. main, master): "
            )
            repo = GitRepo(base_branch=base_branch)
        else:
            input_repo = repo_factory.from_console("repo", allow_none=False)
            assert input_repo is not None
            repo = input_repo

        # Merge approved changes
        steps: List[Step] = []
        if simple or choose_yes_or_no("Automatically merge approved changes?"):
            steps.append(
                ConditionalStep(
                    condition=ReviewStateCondition(
                        comparison=ComparisonType.EQUAL, value=ReviewState.APPROVED
                    ),
                    actions=[MergeAction()],
                )
            )

        # Abandon rejected changes
        if simple or choose_yes_or_no("Automatically abandon rejected changes?"):
            steps.append(
                ConditionalStep(
                    condition=ReviewStateCondition(
                        comparison=ComparisonType.EQUAL, value=ReviewState.CHANGES_REQUESTED
                    ),
                    actions=[AbandonAction()],
                )
            )

        # Update stale changes
        if simple or choose_yes_or_no("Automatically update stale changes?"):
            if simple:
                days_stale = 7
            else:
                days_stale = input_int("Enter number of days until stale", min_val=1)
            steps.append(
                ConditionalStep(
                    condition=UpdatedAgoCondition(
                        comparison=ComparisonType.GREATER_THAN_OR_EQUAL,
                        value=days_stale * 24 * 60 * 60,
                    ),
                    actions=[UpdateAction()],
                )
            )

        return Manager(repo=repo, steps=steps)

    @staticmethod
    def from_console(prev_manager: Optional[Manager] = None) -> Manager:
        """Gets a Manager using console inputs.

        Args:
            prev_manager (Optional[Manager], optional): A previously input Manager.
                Defaults to None.

        Returns:
            Manager: The input Manager.
        """

        args: Dict[str, Any] = {"allow_none": False}
        if prev_manager is not None:
            args["previous_value"] = prev_manager.repo
        repo = repo_factory.from_console("repo", **args)
        assert repo is not None

        if prev_manager is not None and bool(prev_manager.steps):
            steps = choose_options_from_list(
                "Choose steps to keep",
                [(step, f"{step!r}") for step in prev_manager.steps],
                min_choices=0,
                max_choices=len(prev_manager.steps),
            )
        else:
            steps = []
        while choose_yes_or_no("Would you like to add another step?"):
            step = step_factory.from_console("step", allow_none=False)
            assert step is not None
            steps.append(step)

        return Manager(repo=repo, steps=steps)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the GitGrepInput."""

from __future__ import annotations

import subprocess
from typing import ClassVar, Sequence

from autotransform.input.base import Input, InputName
from autotransform.item.file import FileItem


class GitGrepInput(Input):
    """An Input that uses git grep to search a repository for a pattern and returns all files
    that contain a match of the supplied pattern.

    Attributes:
        pattern (str): The pattern to search git grep for.
        name (ClassVar[InputName]): The name of the component.
    """

    pattern: str

    name: ClassVar[InputName] = InputName.GIT_GREP

    def get_items(self) -> Sequence[FileItem]:
        """Gets a list of files using git grep that match the supplied pattern.

        Returns:
            Sequence[FileItem]: The eligible files for transformation.
        """

        git_grep_cmd = [
            "git",
            "grep",
            "--full-name",
            "-l",
            "--untracked",
            "-e",
            self.pattern,
        ]

        try:
            files = subprocess.check_output(git_grep_cmd, encoding="UTF-8").strip().splitlines()
        except subprocess.CalledProcessError:
            return []
        return [FileItem(key=file.replace("\\", "/")) for file in files]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation of RequestAction."""

import os
from functools import cached_property
from typing import Any, ClassVar, Mapping

from pydantic import Field

from autotransform.change.base import Change
from autotransform.step.action.base import Action, ActionName
from autotransform.util.request import RequestHandler


class RequestAction(Action):
    """Performs a URL request for a Change. Used for interacting with REST APIs.

    Attributes:
        url(str): The URL to send a request to.
        data(optional, Mapping[str, Any]): Data to include in the request. Defaults to {}.
        headers(optional, Mapping[str, Any]): Headers to include in the request. Defaults to {}.
        log_response(optional, bool): Indicates whether to log the response using DebugEvent.
            Defaults to False.
        params(optional, Mapping[str, Any]): Params to include in the request. Defaults to {}.
        post(optional, bool): Whether to send the request as a POST. Defaults to True.
        name (ClassVar[ActionName]): The name of the component.
    """

    url: str

    data: Mapping[str, Any] = Field(default_factory=dict)
    headers: Mapping[str, Any] = Field(default_factory=dict)
    log_response: bool = False
    params: Mapping[str, Any] = Field(default_factory=dict)
    post: bool = True

    name: ClassVar[ActionName] = ActionName.REQUEST

    @cached_property
    def _handler(self) -> RequestHandler:
        """Gets the handler for requests.

        Returns:
            RequestHandler: The handler for requests.
        """

        return RequestHandler(
            url=self.url,
            data=self.data,
            headers=self.headers,
            params=self.params,
            log_response=self.log_response,
            post=self.post,
            constant_replacers={"env": lambda var: str(os.getenv(var) or "")},
        )

    def run(self, change: Change) -> bool:
        """Performs a REST API request for a Change.

        Args:
            change (Change): The Change to perform the request on.

        Returns:
            bool: Whether the request returned a non-error response.
        """

        response = self._handler.get_response(
            {"change": lambda name: str(getattr(change, name) or "")}
        )

        return response.ok



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

"""Sets up the build for AutoTransform."""

import os
from typing import List, Tuple

import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


def generate_datafiles() -> List[Tuple[str, List[str]]]:
    """Generates the data files for set-up

    Returns:
        List[Tuple[str, List[str]]]: The datafiles for setup.
    """

    data_files = [
        (
            "autotransform-docs",
            [
                "BEST_PRACTICES.md",
                "COMPONENTS.md",
                "CONTRIBUTING.md",
                "CUSTOM_DEPLOYMENT.md",
                "MANAGE_CHANGES.md",
                "README.md",
                "SCHEDULED_RUNS.md",
            ],
        )
    ]

    for path, _, files in os.walk("examples"):
        data_files.append((f"autotransform-{path}", [os.path.join(path, file) for file in files]))

    return data_files


setuptools.setup(
    name="AutoTransform",
    version="1.1.1a8",
    author="Nathan Rockenbach",
    author_email="nathro.software@gmail.com",
    description="A component based framework for designing automated code modification",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/nathro/AutoTransform",
    project_urls={
        "Source": "https://github.com/nathro/AutoTransform/",
        "Bug Tracker": "https://github.com/nathro/AutoTransform/issues",
    },
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="codemod, automation, code change, codeshift, transformation, maintain",
    package_dir={"": "src/python"},
    packages=setuptools.find_packages("src/python"),
    install_requires=[
        "GitPython==3.1.32",
        "ghapi==1.0.3",
        "typing-extensions==4.4.0",
        "colorama==0.4.6",
        "pytz==2022.7.1",
        "pydantic==1.10.4",
        "libcst==0.4.9",
        "requests==2.31.0",
        "codeowners==0.6.0",
        "openai==0.27.8",
    ],
    python_requires=">=3.10",
    data_files=generate_datafiles(),
    entry_points={
        "console_scripts": [
            "autotransform = autotransform.scripts.main:main",
            "atmigrate-1.0.1 = autotransform.scripts.migrations.p1_0_1:main",
            "atmigrate-1.0.3 = autotransform.scripts.migrations.p1_0_3:main",
            "atmigrate-1.0.5 = autotransform.scripts.migrations.p1_0_5:main",
        ],
    },
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The settings command is used to update AutoTransform settings, such as scheduler.json files,
manager.json files, configs and imported components."""

import json
import os
from argparse import ArgumentParser, Namespace
from typing import Dict, List

from autotransform.config import CONFIG_FILE_NAME, get_cwd_config_dir, get_repo_config_dir
from autotransform.config.config import Config
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.component import ComponentFactory, ComponentImport
from autotransform.util.console import (
    choose_option,
    choose_options_from_list,
    choose_yes_or_no,
    error,
    get_str,
    info,
)
from autotransform.util.enums import SchemaType
from autotransform.util.manager import Manager
from autotransform.util.package import get_config_dir
from autotransform.util.scheduler import Scheduler
from autotransform.util.schema_map import SchemaMap


def add_args(parser: ArgumentParser) -> None:
    """Adds the args to a subparser that are required to update/view settings.

    Args:
        parser (ArgumentParser): The parser for the command.
    """

    setting_type_group = parser.add_mutually_exclusive_group(required=True)
    setting_type_group.add_argument(
        "--user-config",
        dest="setting_type",
        action="store_const",
        const="user_config",
        help="Update or view the user configuration for AutoTransform",
    )
    setting_type_group.add_argument(
        "--repo-config",
        dest="setting_type",
        action="store_const",
        const="repo_config",
        help="Update or view the repo configuration for AutoTransform",
    )
    setting_type_group.add_argument(
        "--cwd-config",
        dest="setting_type",
        action="store_const",
        const="cwd_config",
        help="Update or view the current working directory configuration for AutoTransform",
    )
    setting_type_group.add_argument(
        "--custom-components",
        dest="setting_type",
        action="store_const",
        const="custom_components",
        help="Update or view custom components",
    )
    setting_type_group.add_argument(
        "--manager",
        dest="setting_type",
        action="store_const",
        const="manager",
        help="Update or view manager settings",
    )
    setting_type_group.add_argument(
        "--scheduler",
        dest="setting_type",
        action="store_const",
        const="scheduler",
        help="Update or view scheduler settings",
    )
    setting_type_group.add_argument(
        "--schema-map",
        dest="setting_type",
        action="store_const",
        const="schema_map",
        help="Update or view schema map settings",
    )
    setting_type_group.add_argument(
        "--schema",
        type=str,
        help="The path to an existing or to be created JSON encoded schema.",
    )

    parser.add_argument(
        "--update",
        dest="update_settings",
        action="store_true",
        help="Used to indicate updates are to be made to the settings.",
    )

    parser.set_defaults(func=settings_command_main, update_settings=False)


def settings_command_main(args: Namespace) -> None:
    """The main method for the settings command, handles the actual execution of updating
    and viewing settings.

    Args:
        args (Namespace): The arguments supplied to the settings command.
    """

    if args.setting_type == "user_config":
        path = f"{get_config_dir()}/{CONFIG_FILE_NAME}"
        handle_config(path, "User", args.update_settings)
    elif args.setting_type == "repo_config":
        path = f"{get_repo_config_dir()}/{CONFIG_FILE_NAME}"
        handle_config(path, "Repo", args.update_settings)
    elif args.setting_type == "cwd_config":
        path = f"{get_cwd_config_dir()}/{CONFIG_FILE_NAME}"
        handle_config(path, "CWD", args.update_settings)
    elif args.setting_type == "custom_components":
        handle_custom_components(args.update_settings)
    elif args.setting_type == "manager":
        handle_manager(args.update_settings)
    elif args.setting_type == "scheduler":
        handle_scheduler(args.update_settings)
    elif args.setting_type == "schema_map":
        handle_schema_map(args.update_settings)
    else:
        handle_schema(args.update_settings, args.schema)


def handle_config(path: str, config_type: str, update: bool) -> None:
    """Handles updating a config file.

    Args:
        path (str): The path to the file.
        config_type (str): The type of config being updated (i.e. user).
        update (bool): Whether to update the config.
    """

    config = Config.read(path)
    info(f"Current {config_type} Config\n{config!r}")
    if not update:
        return
    config.from_console(config, user_config=config_type == "User")[0].write(path)


def handle_custom_components(update: bool) -> None:
    """Handle updating/viewing custom components.

    Args:
        update (bool): Whether to apply updates to the custom components.
    """

    component_file_name = choose_options_from_list(
        "Select a component type",
        [
            ("input.json", "Inputs"),
            ("filter.json", "Filters"),
            ("batcher.json", "Batchers"),
            ("transformer.json", "Transformers"),
            ("validator.json", "Validators"),
            ("repo.json", "Repos"),
            ("schema_builder.json", "Schema Builders"),
            ("runner.json", "Runners"),
            ("item.json", "Items"),
            ("change.json", "Changes"),
            ("step.json", "Steps"),
            ("condition.json", "Conditions"),
        ],
    )[0]
    component_dict = ComponentFactory.get_custom_components_dict(component_file_name, strict=False)
    if component_dict:
        info("Custom components:")
        for name, component_import in component_dict.items():
            info(f"\t{name.removeprefix('custom/')}: {component_import!r}")
    else:
        info("No existing custom components")
    if not update:
        return
    # Remove components
    components_to_remove = get_components_to_remove(component_dict)
    changed = bool(components_to_remove)
    for component in components_to_remove:
        del component_dict[component]

    # Add components
    components_to_add = get_components_to_add(component_dict)
    changed = changed or bool(components_to_add)
    component_dict = component_dict | components_to_add

    if changed:
        file_path = ComponentFactory.get_custom_components_path(component_file_name)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+", encoding="UTF-8") as component_file:
            component_file.write(
                json.dumps(
                    {k.removeprefix("custom/"): v.bundle() for k, v in component_dict.items()},
                    indent=4,
                )
            )
            component_file.flush()


def get_components_to_remove(component_dict: Dict[str, ComponentImport]) -> List[str]:
    """Gets a list of components to remove from the dictionary using console input.

    Args:
        component_dict (Dict[str, ComponentImport]): The custom component dictionary.

    Returns:
        List[str]: The keys to remove from the dictionary.
    """

    components = []
    if component_dict:
        name = get_str("Enter a component name to remove(blank to skip): ")
    else:
        name = ""
    while name != "":
        if name.startswith("custom/"):
            name = name.removeprefix("custom/")
        if f"custom/{name}" not in component_dict:
            error(f"No component import with name: {name}")
        elif f"custom/{name}" in components:
            error(f"Already removing component import with name: {name}")
        else:
            components.append(f"custom/{name}")
        if len(component_dict) <= len(components):
            break
        name = get_str("Enter a component name to remove(blank to skip): ")
    return components


def get_components_to_add(component_dict: Dict[str, ComponentImport]) -> Dict[str, ComponentImport]:
    """Gets a dictionary of new components to add to the custom component imports.

    Args:
        component_dict (Dict[str, ComponentImport]): The existing custom components.

    Returns:
        Dict[str, ComponentImport]: The components to add to the dictionary.
    """

    components_to_add = {}
    name = get_str("Enter component name to add(blank to skip): ")
    while name != "":
        if name.startswith("custom/"):
            name = name.removeprefix("custom/")
        if f"custom/{name}" in component_dict:
            error(f"Component already exists with name: {name}")
        elif f"custom/{name}" in components_to_add:
            error(f"Already adding component with name: {name}")
        else:
            class_name = get_str("Enter the class representing this component: ")
            module = get_str("Enter the fully qualified name of the module for the class: ")
            components_to_add[f"custom/{name}"] = ComponentImport(
                class_name=class_name, module=module
            )
        name = get_str("Enter component name to add(blank to skip): ")

    return components_to_add


def handle_manager(update: bool) -> None:
    """Handle updating/viewing the Manager.

    Args:
        update (bool): Whether to apply updates to the Manager.
    """

    path = f"{get_repo_config_dir()}/manager.json"
    manager = None
    try:
        manager = Manager.read(path)
        info(f"Current Manager\n{manager!r}")
    except FileNotFoundError:
        error("No Manager file found")
    except json.JSONDecodeError as err:
        error(f"Failed to decode Manager JSON\n{err}")
    except ValueError as err:
        error(f"Invalid Manager value\n{err}")
    except TypeError as err:
        error(f"Invalid Manager type\n{err}")

    if not update:
        return
    Manager.from_console(manager).write(path)


def handle_scheduler(update: bool) -> None:
    """Handle updating/viewing the Scheduler.

    Args:
        update (bool): Whether to apply updates to the Scheduler.
    """

    path = f"{get_repo_config_dir()}/scheduler.json"
    scheduler = None
    try:
        scheduler = Scheduler.read(path)
        info(f"Current Scheduler\n{scheduler!r}")
    except FileNotFoundError:
        error("No Manager file found")
    except json.JSONDecodeError as err:
        error(f"Failed to decode Manager JSON\n{err}")
    except ValueError as err:
        error(f"Invalid Manager value\n{err}")
    except TypeError as err:
        error(f"Invalid Manager type\n{err}")

    if not update:
        return
    Scheduler.from_console(scheduler).write(path)


# pylint: disable=too-many-return-statements,too-many-branches
def handle_schema_map(update: bool) -> None:
    """Handle updating/viewing the Schema Map.

    Args:
        update (bool): Whether to apply updates to the Schema Map.
    """

    schema_map = SchemaMap.get()

    info("Current Schema Map")
    for schema_name, schema_info in schema_map.items():
        info(f"{schema_name}: {schema_info[1]} ({schema_info[0]})")

    if not update:
        return

    schemas = choose_options_from_list(
        "Choose Schemas to keep",
        [(name, name) for name, _ in schema_map.items()],
        min_choices=0,
        max_choices=len(schema_map.items()),
    )
    schemas_to_remove = [schema for schema, _ in schema_map.items() if schema not in schemas]
    for schema in schemas_to_remove:
        schema_map.remove_schema(schema)

    while choose_yes_or_no("Would you like to add a Schema to the map?"):
        name = get_str("Schema Name: ")
        while name in schema_map:  # pylint: disable=unsupported-membership-test
            error(f"{name} already present in the Schema Map")
            name = get_str("Schema Name: ")
        schema_type = choose_option(
            "Choose SchemaType",
            [(SchemaType.BUILDER, ["builder", "b"]), (SchemaType.FILE, ["file", "f"])],
        )
        if schema_type == SchemaType.BUILDER:
            target = get_str("Enter SchemaBuilder class: ")
        else:
            target = get_str("Enter Schema filepath: ")
        schema_map.add_schema(name, schema_type, target)

    schema_map.write()


def handle_schema(update: bool, file_path: str) -> None:
    """Handle updating/viewing a Schema.

    Args:
        update (bool): Whether to apply updates to the Schema.
        file_path (str): The path to the Schema file.
    """
    try:
        with open(file_path, "r", encoding="UTF-8") as schema_file:
            schema = AutoTransformSchema.from_data(json.loads(schema_file.read()))
        info(f"Existing schema:\n{schema!r}")
    except FileNotFoundError:
        info(f"No schema found at path: {file_path}")
        schema = None

    if update:
        schema = AutoTransformSchema.from_console(schema)
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+", encoding="UTF-8") as schema_file:
            schema_file.write(json.dumps(schema.bundle(), indent=4))
            schema_file.flush()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Utility functions used by multiple components."""

import json
import os
import subprocess
from tempfile import NamedTemporaryFile as TmpFile
from typing import Any, Dict, List, Mapping, Optional, Sequence

from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.item.base import Item


def run_cmd_on_items(
    cmd: List[str],
    items: Sequence[Item],
    batch_metadata: Mapping[str, Any],
    timeout: Optional[int] = None,
) -> subprocess.CompletedProcess:
    """Creates and runs a subprocess based on supplied command and replacements.

    Args:
        cmd (List[str]): The command to run.
        items (Sequence[Item]): The items that are being run on.
        batch_metadata (Mapping[str, Any]): The metadata for the batch.
        timeout (optional, Optional[int]): A timeout for the subprocess run. Defaults to None.

    Returns:
        subprocess.CompletedProcess: The completed process run.
    """

    event_handler = EventHandler.get()
    item_keys = [item.key for item in items]
    extra_data = {item.key: item.extra_data for item in items if item.extra_data is not None}
    arg_replacements = {
        "<<KEY>>": item_keys,
        "<<EXTRA_DATA>>": [json.dumps(extra_data)],
        "<<METADATA>>": [json.dumps(batch_metadata)],
    }

    with TmpFile(mode="w+") as inp, TmpFile(mode="w+") as meta, TmpFile(mode="w+") as extra:
        # Make key file
        json.dump(item_keys, inp)
        inp.flush()
        arg_replacements["<<KEY_FILE>>"] = [inp.name]

        # Make extra_data file
        json.dump(extra_data, extra)
        extra.flush()
        arg_replacements["<<EXTRA_DATA_FILE>>"] = [extra.name]

        # Make metadata file
        json.dump(batch_metadata, meta)
        meta.flush()
        arg_replacements["<<METADATA_FILE>>"] = [meta.name]

        # Create command
        replaced_cmd = replace_script_args(cmd, arg_replacements)

        # Run script
        event_handler.handle(VerboseEvent({"message": f"Running command: {replaced_cmd}"}))
        return subprocess.run(
            replaced_cmd, capture_output=True, encoding="utf-8", check=False, timeout=timeout
        )


def replace_script_args(args: List[str], replacements: Dict[str, List[str]]) -> List[str]:
    """Replaces arguments in a list with replacements that are supplied.

    Args:
        args (List[str]): The argument list.
        replacements (Dict[str, Tuple[List[str], bool]]): Maps a value to the list of args that
            replaces it.

    Returns:
        List[str]: The list of args with replacements.
    """

    script_replacements = os.getenv("AUTO_TRANSFORM_SCRIPT_REPLACEMENTS")
    if script_replacements:
        env_replacements: Dict[str, List[str]] = json.loads(script_replacements)
    else:
        env_replacements = {}

    replaced_args = []
    for arg in args:
        if arg in replacements:
            replaced_args.extend(replacements[arg])
        elif arg in env_replacements:
            replaced_args.extend(env_replacements[arg])
        else:
            replaced_args.append(arg)
    return replaced_args



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Condition components."""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, ClassVar, Dict, Generic, List, Optional, Set, Type, TypeVar

from pydantic import root_validator, validator

from autotransform.change.base import Change
from autotransform.step.condition.comparison import ComparisonType, compare
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class ConditionName(str, Enum):
    """A simple enum for mapping."""

    AGGREGATE = "aggregate"
    CHANGE_STATE = "change_state"
    CREATED_AGO = "created_ago"
    LABELS = "labels"
    MERGEABLE_STATE = "mergeable_state"
    REQUEST_STR = "request_str"
    REVIEW_STATE = "review_state"
    REVIEWERS = "reviewers"
    SCHEMA_NAME = "schema_name"
    TEAM_REVIEWERS = "team_reviewers"
    TEST_STATE = "test_state"
    UPDATED_AGO = "updated_ago"


class Condition(NamedComponent):
    """The base for Condition components. Used by ConditionalStep to determine whether to
    take an Action.

    Attributes:
        name (ClassVar[ConditionName]): The name of the Component.
    """

    name: ClassVar[ConditionName]

    @abstractmethod
    def check(self, change: Change) -> bool:
        """Checks whether the Change passes the Condition.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            bool: Whether the Change passes the condition.
        """


T = TypeVar("T")


class ComparisonCondition(Generic[T], Condition):
    """The base for comparison Condition components. Uses the ComparisonType enum to perform
    comparisons.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (T | List[T]): The value to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: T | List[T]

    name: ClassVar[ConditionName]

    @abstractmethod
    def get_val_from_change(self, change: Change) -> T:
        """Gets the existing value to compare against from the change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            T: The value from the Change to compare against.
        """

    def check(self, change: Change) -> bool:
        """Checks whether the Change passes the Condition.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            bool: Whether the Change passes the condition.
        """

        return compare(self.get_val_from_change(change), self.value, self.comparison)

    @staticmethod
    def valid_comparisons() -> Set[ComparisonType]:
        """Gets the valid comparisons that can be done for this condition.

        Returns:
            Set[ComparisonType]: The valid comparisons this condition can perform.
        """

        return {
            ComparisonType.EQUAL,
            ComparisonType.NOT_EQUAL,
            ComparisonType.IN,
            ComparisonType.NOT_IN,
        }

    @root_validator
    @classmethod
    def check_value_for_comparison(
        cls: Type[ComparisonCondition], values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Ensures that the value is provided as either a single value or a list, depending
        on the comparison type.

        Args:
            cls (Type[ListComparisonCondition]): The Condition class.
            values (Dict[str, Any]): The values for the comparison.

        Raises:
            ValueError: Raises errors when value provided does not work with the

        Returns:
            Dict[str, Any]: The unmodified values for comparison.
        """

        comparison = values["comparison"]
        if comparison in [ComparisonType.IN, ComparisonType.NOT_IN] and not isinstance(
            values["value"], List
        ):
            raise ValueError(f"Can not perform comparison {comparison} when value is not a list.")

        if comparison not in [ComparisonType.IN, ComparisonType.NOT_IN] and isinstance(
            values["value"], List
        ):
            raise ValueError(f"Can not perform comparison {comparison} when value is a list.")

        return values

    # pylint: disable=invalid-name
    @validator("comparison")
    @classmethod
    def comparison_type_is_valid(
        cls: Type[ComparisonCondition], v: ComparisonType
    ) -> ComparisonType:
        """Validates the condition can perform the specified comparison.

        Args:
            cls (Type[ComparisonCondition]): The Condition class.
            v (ComparisonType): The comparison to perform.

        Raises:
            ValueError: Raised if the Condition can not perform the comparison.

        Returns:
            ComparisonType: The unmodified comparison to perform.
        """

        if v not in cls.valid_comparisons():
            raise ValueError(f"{cls.__name__} can not perform comparison {v}")
        return v


class SortableComparisonCondition(ABC, Generic[T], ComparisonCondition[T]):
    """The base for sortable comparison Condition components. Uses the ComparisonType enum to
    perform comparisons that are sortable.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (T): The value to compare against.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    @staticmethod
    def valid_comparisons() -> Set[ComparisonType]:
        """Gets the valid comparisons that can be done for this condition.

        Returns:
            Set[ComparisonType]: The valid comparisons this condition can perform.
        """

        return {
            ComparisonType.EQUAL,
            ComparisonType.NOT_EQUAL,
            ComparisonType.IN,
            ComparisonType.NOT_IN,
            ComparisonType.GREATER_THAN,
            ComparisonType.GREATER_THAN_OR_EQUAL,
            ComparisonType.LESS_THAN,
            ComparisonType.LESS_THAN_OR_EQUAL,
        }


class ListComparisonCondition(Generic[T], Condition):
    """The base for sortable comparison Condition components. Uses the ComparisonType enum to
    perform comparisons that are sortable.

    Attributes:
        comparison (ComparisonType): The type of comparison to perform.
        value (optional, Optional[T]): The value to compare against. Defaults to None.
        name (ClassVar[ConditionName]): The name of the Component.
    """

    comparison: ComparisonType
    value: Optional[T] = None

    name: ClassVar[ConditionName]

    @abstractmethod
    def get_val_from_change(self, change: Change) -> List[T]:
        """Gets the existing value to compare against from the change.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            List[T]: The value from the Change to compare against.
        """

    def check(self, change: Change) -> bool:
        """Checks whether the Change passes the Condition.

        Args:
            change (Change): The Change the Condition is checking.

        Returns:
            bool: Whether the Change passes the condition.
        """

        return compare(self.get_val_from_change(change), self.value, self.comparison)

    @root_validator
    @classmethod
    def check_value_for_comparison(
        cls: Type[ListComparisonCondition], values: Dict[str, Any]
    ) -> Dict[str, Any]:
        """Ensures that the value is provided for contains/not contains checks
        or not provided for empty/not empty checks.

        Args:
            cls (Type[ListComparisonCondition]): The Condition class.
            values (Dict[str, Any]): The values for the comparison.

        Raises:
            ValueError: Raises errors when values are provided for an empty check
                or not provided for a contains check.

        Returns:
            Dict[str, Any]: The unmodified values for comparison.
        """

        comparison = values["comparison"]
        if (
            comparison in [ComparisonType.EMPTY, ComparisonType.NOT_EMPTY]
            and values.get("value") is not None
        ):
            raise ValueError(f"Can not perform comparison {comparison} when value is not None.")
        if (
            comparison in [ComparisonType.CONTAINS, ComparisonType.NOT_CONTAINS]
            and values.get("value") is None
        ):
            raise ValueError(f"Can not perform comparison {comparison} when value is None.")

        return values

    @staticmethod
    def valid_comparisons() -> Set[ComparisonType]:
        """Gets the valid comparisons that can be done for this condition.

        Returns:
            Set[ComparisonType]: The valid comparisons this condition can perform.
        """

        return {
            ComparisonType.CONTAINS,
            ComparisonType.NOT_CONTAINS,
            ComparisonType.EMPTY,
            ComparisonType.NOT_EMPTY,
        }

    # pylint: disable=invalid-name
    @validator("comparison")
    @classmethod
    def comparison_type_is_valid(
        cls: Type[ListComparisonCondition], v: ComparisonType
    ) -> ComparisonType:
        """Validates the condition can perform the specified comparison.

        Args:
            cls (Type[ListComparisonCondition]): The Condition class.
            v (ComparisonType): The comparison to perform.

        Raises:
            ValueError: Raised if the Condition can not perform the comparison.

        Returns:
            ComparisonType: The unmodified comparison to perform.
        """

        if v not in cls.valid_comparisons():
            raise ValueError(f"{cls.__name__} can not perform comparison {v}")
        return v


FACTORY = ComponentFactory(
    {
        ConditionName.AGGREGATE: ComponentImport(
            class_name="AggregateCondition", module="autotransform.step.condition.aggregate"
        ),
        ConditionName.CHANGE_STATE: ComponentImport(
            class_name="ChangeStateCondition", module="autotransform.step.condition.state"
        ),
        ConditionName.CREATED_AGO: ComponentImport(
            class_name="CreatedAgoCondition", module="autotransform.step.condition.created"
        ),
        ConditionName.LABELS: ComponentImport(
            class_name="LabelsCondition", module="autotransform.step.condition.labels"
        ),
        ConditionName.MERGEABLE_STATE: ComponentImport(
            class_name="MergeableStateCondition", module="autotransform.step.condition.state"
        ),
        ConditionName.REQUEST_STR: ComponentImport(
            class_name="RequestStrCondition", module="autotransform.step.condition.request"
        ),
        ConditionName.REVIEW_STATE: ComponentImport(
            class_name="ReviewStateCondition", module="autotransform.step.condition.state"
        ),
        ConditionName.REVIEWERS: ComponentImport(
            class_name="ReviewersCondition", module="autotransform.step.condition.reviewers"
        ),
        ConditionName.SCHEMA_NAME: ComponentImport(
            class_name="SchemaNameCondition", module="autotransform.step.condition.schema"
        ),
        ConditionName.TEAM_REVIEWERS: ComponentImport(
            class_name="TeamReviewersCondition", module="autotransform.step.condition.reviewers"
        ),
        ConditionName.TEST_STATE: ComponentImport(
            class_name="TestStateCondition", module="autotransform.step.condition.state"
        ),
        ConditionName.UPDATED_AGO: ComponentImport(
            class_name="UpdatedAgoCondition", module="autotransform.step.condition.updated"
        ),
    },
    Condition,  # type: ignore [type-abstract]
    "condition.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Provides utility functions for traversing the overall package."""

import pathlib


def get_path_as_str(path: pathlib.Path) -> str:
    """Coverts a path to a string AutoTransform can use.

    Args:
        path (pathlib.Path): The path to convert.

    Returns:
        str: The path as a string.
    """

    return str(path.resolve()).replace("\\", "/")


def get_package_dir() -> pathlib.Path:
    """Gets the directory for the autotransform package

    Returns:
        pathlib.Path: The package's directory.
    """

    path = pathlib.Path(__file__)
    for parent in path.parents:
        if (parent / "examples").exists() or (parent / "autotransform-examples").exists():
            return parent
    while path.name != "autotransform":
        path = path.parent
    return path.parent


def get_examples_dir() -> str:
    """Gets the directory where examples are located.

    Returns:
        str: The examples directory
    """

    package_dir = get_package_dir()
    examples_dir = package_dir / "examples"
    if examples_dir.exists():
        return get_path_as_str(examples_dir)
    examples_dir = package_dir / "autotransform-examples"
    if examples_dir.exists():
        return get_path_as_str(examples_dir)
    raise FileNotFoundError()


def get_config_dir() -> str:
    """Gets the directory where the configuration is stored.

    Returns:
        str: The directory where configuration is stored.
    """
    package_dir = get_package_dir()
    if (package_dir / "examples").exists():
        config_dir = "config"
    else:
        config_dir = "autotransform-config"
    return get_path_as_str(get_package_dir() / config_dir)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Runner components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import ClassVar

from autotransform.change.base import Change
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class RunnerName(str, Enum):
    """A simple enum for mapping."""

    GITHUB = "github"
    JENKINS_API = "jenkins_api"
    JENKINS_FILE = "jenkins_file"
    LOCAL = "local"


class Runner(NamedComponent):
    """The base for Runner components. Used by AutoTransform to run an AutoTransformSchema,
    either locally on the machine or on remote infrastructure.

    Attributes:
        name (ClassVar[RunnerName]): The name of the component.
    """

    name: ClassVar[RunnerName]

    @abstractmethod
    def run(self, schema: AutoTransformSchema) -> None:
        """Triggers a full run of a Schema.

        Args:
            schema (AutoTransformSchema): The schema that will be run.
        """

    @abstractmethod
    def update(self, change: Change) -> None:
        """Triggers an update of the Change.

        Args:
            change (Change): The Change to update.
        """


FACTORY = ComponentFactory(
    {
        RunnerName.GITHUB: ComponentImport(
            class_name="GithubRunner", module="autotransform.runner.github"
        ),
        RunnerName.JENKINS_API: ComponentImport(
            class_name="JenkinsAPIRunner", module="autotransform.runner.jenkins"
        ),
        RunnerName.JENKINS_FILE: ComponentImport(
            class_name="JenkinsFileRunner", module="autotransform.runner.jenkins"
        ),
        RunnerName.LOCAL: ComponentImport(
            class_name="LocalRunner", module="autotransform.runner.local"
        ),
    },
    Runner,  # type: ignore [type-abstract]
    "runner.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Item components. Can be used as a generic Item."""

from __future__ import annotations

from enum import Enum
from typing import Any, ClassVar, Dict, Optional

from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class ItemName(str, Enum):
    """A simple enum for mapping."""

    FILE = "file"
    GENERIC = "generic"


class Item(NamedComponent):
    """The base for Item components. Returned by Input components and act as the units on which
    AutoTransform operates.

    Attributes:
        key (str): A unique key used to represent this item, such as a file path.
        extra_data (Optional[Dict[str, Any]], optional): Any extra data that needs to be
            associated with this Item. Should be JSON encodable. Defaults to None.
        name (ClassVar[ItemName]): The name of the component.
    """

    key: str
    extra_data: Optional[Dict[str, Any]] = None

    name: ClassVar[ItemName] = ItemName.GENERIC


FACTORY = ComponentFactory(
    {
        ItemName.FILE: ComponentImport(class_name="FileItem", module="autotransform.item.file"),
        ItemName.GENERIC: ComponentImport(class_name="Item", module="autotransform.item.base"),
    },
    Item,  # type: ignore [misc]
    "item.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the JSCodeshiftTransformer. See https://github.com/facebook/jscodeshift"""

from __future__ import annotations

import subprocess
from typing import ClassVar, List

from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.item.base import Item
from autotransform.item.file import FileItem
from autotransform.transformer.base import TransformerName
from autotransform.transformer.single import SingleTransformer
from pydantic import Field


class JSCodeshiftTransformer(SingleTransformer):
    """A Transformer that makes changes using JSCodeshift.

    Attributes:
        js_transform (str): The JSCodeshift transform to execute.
        args (optional, List[str]): The arguments to supply to the transformation. Defaults to [].
        timeout (optional, int): The timeout for an individual run.
        name (ClassVar[TransformerName]): The name of the component.
    """

    js_transform: str

    args: List[str] = Field(default_factory=list)
    timeout: int = 600

    name: ClassVar[TransformerName] = TransformerName.JSCODESHIFT

    def _transform_item(self, item: Item) -> None:
        """Run the supplied JSCodeshift transform against the file.

        Args:
            item (Item): The file that will be transformed.
        """

        assert isinstance(item, FileItem)
        event_handler = EventHandler.get()

        cmd = ["jscodeshift", "-t", self.js_transform, item.get_path()]
        cmd.extend(self.args)

        # Run JSCodeshift
        event_handler.handle(VerboseEvent({"message": f"Running command: {cmd}"}))
        proc = subprocess.run(
            cmd,
            capture_output=True,
            encoding="utf-8",
            check=False,
            timeout=self.timeout,
        )
        if proc.stdout.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDOUT"}))
        if proc.stderr.strip() != "":
            event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
        else:
            event_handler.handle(VerboseEvent({"message": "No STDERR"}))
        proc.check_returncode()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""A script to perform a migration of a Manager JSON file from 1.0.0 to 1.0.1."""

import json
from argparse import ArgumentParser
from typing import Any, Dict

from autotransform.config import get_repo_config_relative_path
from autotransform.schema.builder import FACTORY as schema_builder_factory
from autotransform.schema.schema import AutoTransformSchema
from autotransform.util.enums import SchemaType
from autotransform.util.scheduler import Scheduler
from autotransform.util.schema_map import SchemaMap


def get_arg_parser() -> ArgumentParser:
    """Gets the argument parser for migrating the Scheduler JSON file.

    Returns:
        ArgumentParser: The arg parser with all args set up.
    """

    parser = ArgumentParser(
        description="Upgrade Scheduler JSON files for 1.0.2 -> 1.0.3",
        prog="AutoTransform",
    )

    parser.add_argument(
        "--path",
        metavar="path",
        required=False,
        type=str,
        help="A file path to the JSON encoded file, only use if file is in a non-usual place.",
    )

    return parser


def main() -> None:
    """Migrate Scheduler and Schema map for 1.0.2 -> 1.0.3."""

    parser = get_arg_parser()
    args = parser.parse_args()

    # Get existing Scheduler data
    file_path = args.path
    if file_path is None:
        file_path = f"{get_repo_config_relative_path()}/scheduler.json"

    with open(file_path, "r", encoding="UTF-8") as scheduler_file:
        scheduler_json = scheduler_file.read()

    scheduler_data = json.loads(scheduler_json)

    schema_map = SchemaMap.get()
    update_scheduler_data(scheduler_data, schema_map)

    scheduler = Scheduler.from_data(scheduler_data)
    scheduler.write(file_path)
    schema_map.write()


def update_scheduler_data(scheduler_data: Dict[str, Any], schema_map: SchemaMap) -> None:
    """Updates Scheduler data and the Schema map for 1.0.2 -> 1.0.3 conversion.

    Args:
        scheduler_data (Dict[str, Any]): The existing Scheduler data.
        schema_map (SchemaMap): The existing Schema map.
    """

    for schema_data in scheduler_data["schemas"]:
        schema_type = schema_data.get("type")

        # If type isn't present, it was already converted
        if schema_type is None:
            continue
        schema_type = SchemaType(schema_type)
        schema_target = schema_data["target"]
        if schema_type == SchemaType.BUILDER:
            schema = schema_builder_factory.get_instance({"name": schema_target}).build()
        else:
            with open(schema_target, "r", encoding="UTF-8") as schema_file:
                schema = AutoTransformSchema.from_data(json.loads(schema_file.read()))
        schema_name = schema.config.schema_name
        del schema_data["target"]
        del schema_data["type"]
        schema_data["schema_name"] = schema_name
        if schema_name not in schema_map:
            schema_map.add_schema(schema_name, schema_type, schema_target)


if __name__ == "__main__":
    main()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the SingleBatcher."""

from __future__ import annotations

from copy import deepcopy
from typing import Any, ClassVar, Dict, List, Optional, Sequence

from autotransform.batcher.base import Batch, Batcher, BatcherName
from autotransform.item.base import Item


class SingleBatcher(Batcher):
    """A batcher which puts all Items together in to a single batch.

    Attributes:
        title (str): The title to use for the Batch.
        metadata (Optional[Dict[str, Any]], optional): The metadata to use for the
            Batch. Defaults to None.
        skip_empty_batch (Optional[bool], optional): Whether to skip Batch creation if
            there are no Items. Defaults to False.
        name (ClassVar[BatcherName]): The name of the Component.
    """

    title: str
    metadata: Optional[Dict[str, Any]] = None
    skip_empty_batch: bool = False

    name: ClassVar[BatcherName] = BatcherName.SINGLE

    def batch(self, items: Sequence[Item]) -> List[Batch]:
        """Takes in a list Items and batches them together in to a single Batch.

        Args:
            items (Sequence[Item]): The filtered Items to separate.

        Returns:
            List[Batch]: A list containing a single Batch for all Items.
        """

        # Skip if empty when setting is enabled
        if self.skip_empty_batch and len(items) == 0:
            return []

        # Create Batch
        batch: Batch = {
            "items": items,
            "title": self.title,
        }

        # Deepcopy metadata to ensure mutations don't apply to params
        if self.metadata is not None:
            batch["metadata"] = deepcopy(self.metadata)
        return [batch]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Repo components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import Any, ClassVar, List, Mapping, Optional, Sequence

from autotransform.batcher.base import Batch
from autotransform.change.base import Change
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class RepoName(str, Enum):
    """A simple enum for mapping."""

    GIT = "git"
    GITHUB = "github"


class Repo(NamedComponent):
    """The base for Repo components. Used by AutoTransform to interact with version
    control and code review systems.

    Attributes:
        name (ClassVar[RepoName]): The name of the component.
    """

    name: ClassVar[RepoName]

    @abstractmethod
    def get_changed_files(self, batch: Batch) -> List[str]:
        """Gets all files in the repo that have been modified.

        Args:
            batch (Batch): The Batch that was used for the transformation

        Returns:
            List[str]: The list of files that have unsubmitted changes.
        """

    def has_changes(self, batch: Batch) -> bool:
        """Check whether any changes have been made to the underlying code based on the Batch.

        Args:
            batch (Batch): The Batch that was used for the transformation.

        Returns:
            bool: Returns True if there are any changes to the codebase.
        """

        return len(self.get_changed_files(batch)) > 0

    @abstractmethod
    def has_outstanding_change(self, batch: Batch) -> bool:
        """Checks the state of the repo to see whether an outstanding Change
        for the Batch exists.

        Args:
            batch (Batch): The Batch to check for.

        Returns:
            bool: Whether an outstanding Change exists.
        """

    @abstractmethod
    def submit(
        self,
        batch: Batch,
        transform_data: Optional[Mapping[str, Any]],
        change: Optional[Change] = None,
    ) -> None:
        """Submit the changes to the Repo (i.e. commit, submit Pull Request, etc...).
        Only called when changes are present.

        Args:
            batch (Batch): The Batch for which the changes were made.
            transform_data (Optional[Mapping[str, Any]]): Data from the transformation.
            change (Optional[Change]): An associated change which should be updated.
                Defaults to None.
        """

    @abstractmethod
    def clean(self, batch: Batch) -> None:
        """Clean any changes present in the Repo that have not been submitted.

        Args:
            batch (Batch): The Batch for which we are cleaning the repo.
        """

    @abstractmethod
    def rewind(self, batch: Batch) -> None:
        """Rewind the repo to a pre-submit state to prepare for executing another Batch. This
        should NOT delete any submissions (i.e. commits should stay present).

        Args:
            batch (Batch): The Batch for which changes were submitted.
        """

    @abstractmethod
    def get_outstanding_changes(self) -> Sequence[Change]:
        """Gets all outstanding Changes for the Repo.

        Returns:
            Sequence[Change]: The outstanding Changes against the Repo.
        """


FACTORY = ComponentFactory(
    {
        RepoName.GIT: ComponentImport(class_name="GitRepo", module="autotransform.repo.git"),
        RepoName.GITHUB: ComponentImport(
            class_name="GithubRepo", module="autotransform.repo.github"
        ),
    },
    Repo,  # type: ignore [type-abstract]
    "repo.json",
)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for script based Inputs."""

from __future__ import annotations

import json
import subprocess
from tempfile import NamedTemporaryFile
from typing import ClassVar, List, Sequence

from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.input.base import Input, InputName
from autotransform.item.base import FACTORY as item_factory
from autotransform.item.base import Item
from autotransform.util.functions import replace_script_args


class ScriptInput(Input):
    """An Input that uses a script to generate a list of Items.

    Attributes:
        args (List[str]): The arguments to supply to the script.
        script (str): The script to run.
        timeout (int): The timeout to use for the script process.
        name (ClassVar[InputName]): The name of the component.
    """

    args: List[str]
    script: str
    timeout: int

    name: ClassVar[InputName] = InputName.SCRIPT

    def get_items(self) -> Sequence[Item]:
        """Uses a simple script to generate a list of Items. If a <<RESULT_FILE>> arg is used
        it will be replaced with the path of a temporary file that can be used to store a JSON
        encoded list of Items that will be returned. If no such arg is used, the STDOUT of the
        script will be interpreted as a JSON encoded list of Items.

        Returns:
            Sequence[Item]: The supplied Items.
        """

        event_handler = EventHandler.get()

        # Get Command
        cmd = [self.script]
        cmd.extend(self.args)

        with NamedTemporaryFile(mode="r+b") as result_file:
            arg_replacements = {"<<RESULT_FILE>>": [result_file.name]}
            uses_result_file = "<<RESULT_FILE>>" in cmd
            replaced_cmd = replace_script_args(cmd, arg_replacements)

            # Run script
            event_handler.handle(VerboseEvent({"message": f"Running command: {replaced_cmd}"}))
            proc = subprocess.run(
                replaced_cmd,
                capture_output=True,
                encoding="utf-8",
                check=False,
                timeout=self.timeout,
            )

            if proc.stdout.strip() != "" and uses_result_file:
                event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
            elif uses_result_file:
                event_handler.handle(VerboseEvent({"message": "No STDOUT"}))

            if proc.stderr.strip() != "":
                event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
            else:
                event_handler.handle(VerboseEvent({"message": "No STDERR"}))
            proc.check_returncode()
            if uses_result_file:
                with open(result_file.name, encoding="utf-8") as results:
                    item_data = json.loads(results.read())
            else:
                item_data = json.loads(proc.stdout.strip())
        return [item_factory.get_instance(item) for item in item_data]



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""All Actions associated with a Change's comments."""

from typing import ClassVar

from pydantic import validator

from autotransform.change.base import Change
from autotransform.step.action.base import Action, ActionName


class CommentAction(Action):
    """Adds a comment to an existing Change.

    Attributes:
        body(str): The body of the comment.
        name (ClassVar[ActionName]): The name of the component.
    """

    body: str

    name: ClassVar[ActionName] = ActionName.COMMENT

    # pylint: disable=invalid-name
    @validator("body")
    @classmethod
    def body_must_be_non_empty(cls, v: str) -> str:
        """Validates the body is not empty.

        Args:
            v (str): The body of the comment.

        Raises:
            ValueError: Raises an error when the body is empty.

        Returns:
            str: The unmodified body of the comment.
        """

        if v == "":
            raise ValueError("Comment body must be non-empty")
        return v

    def run(self, change: Change) -> bool:
        """Adds a comment to the specified Change.

        Args:
            change (Change): The Change to comment on.

        Returns:
            bool: Whether the comment was successful.
        """

        return change.comment(self.body)



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The implementation for the ScriptTransformer."""

from __future__ import annotations

from typing import Any, ClassVar, Dict, List, Optional

from autotransform.batcher.base import Batch
from autotransform.event.handler import EventHandler
from autotransform.event.verbose import VerboseEvent
from autotransform.transformer.base import Transformer, TransformerName
from autotransform.util.functions import run_cmd_on_items
from pydantic import root_validator, validator


class ScriptTransformer(Transformer[None]):
    """A Transformer that makes changes using an invoked script. Sentinel values can be
    used in args to provide custom arguments for a run.
    The available sentinel values for args are:
        <<KEY>>: A json encoded list of the Items for a Batch. If the per_item flag is set
            this will simply be the key of an Item.
        <<EXTRA_DATA>>: A JSON encoded mapping from Item key to that Item's extra_data. If the
            per_item flag is set, this will simply be a JSON encoding of the Item's extra_data.
            If extra_data is not present for an item, it is treated as an empty Dict.
        <<METADATA>>: A JSON encoded version of the Batch's metadata.
    _FILE can be appended to any of these (i.e. <<KEY_FILE>>) and the arg will instead be replaced
     with a path to a file containing the value.

    Attributes:
        args (List[str]): The arguments to supply to the script.
        script (str): The script to run.
        timeout (int): The timeout to use for the script process.
        chunk_size (Optional[int], optional): The size of chunks to operate on. None indicates no
            chunking. Defaults to None.
        name (ClassVar[TransformerName]): The name of the Component.
    """

    args: List[str]
    script: str
    timeout: int
    chunk_size: Optional[int] = None

    name: ClassVar[TransformerName] = TransformerName.SCRIPT

    # pylint: disable=invalid-name
    @validator("chunk_size")
    @classmethod
    def chunk_size_must_be_positive(cls, v: Optional[int]) -> Optional[int]:
        """Validates that chunk

        Args:
            v (Optional[int]): The chunk_size that was set.

        Raises:
            ValueError: Raised if chunk_size is not None and is less than 1.

        Returns:
            Optional[int]: The validated chunk_size.
        """

        if v is not None and v < 1:
            raise ValueError(f"Chunk size must be greater than 0, {v} provided")
        return v

    @root_validator(pre=True)
    @classmethod
    def per_item_legacy_setting_validator(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        """Validates chunk size using legacy per_item setting.

        Args:
            values (Dict[str, Any]): The values used to configure the ScriptTransformer.

        Raises:
            ValueError: Raised if per item is set with a chunk_size that is not 1.

        Returns:
            Mapping[str, Any]: The fixed values.
        """

        if "per_item" in values and values["per_item"]:
            if "chunk_size" in values and values["chunk_size"] != 1:
                raise ValueError("Per item can not be specified with a chunk size that is not 1")
            values["chunk_size"] = 1
            return values
        return values

    def transform(self, batch: Batch) -> None:
        """Executes a simple script to transform the given Batch. Sentinel values can be
        used in args to provide custom arguments for a run.
        The available sentinel values for args are:
            <<KEY>>: A json encoded list of the Items for a Batch.
            <<EXTRA_DATA>>: A JSON encoded mapping from Item key to that Item's extra_data.
                If extra_data is not present for an item, it is treated as an empty Dict.
            <<METADATA>>: A JSON encoded version of the Batch's metadata.
        _FILE can be appended to any of these (i.e. <<KEY_FILE>>) and the arg will instead be
        replaced with a path to a file containing the value.

        Args:
            batch (Batch): The batch that will be transformed.
        """

        event_handler = EventHandler.get()

        metadata = batch.get("metadata", {})
        items = batch["items"]
        num_items = max(len(items), 1)
        chunk_size = self.chunk_size or num_items

        # Get Command
        for i in range(0, num_items, chunk_size):
            chunk_items = items[i : i + chunk_size]
            cmd = [self.script]
            cmd.extend(self.args)

            proc = run_cmd_on_items(cmd, chunk_items, metadata, timeout=self.timeout)

            if proc.stdout.strip() != "":
                event_handler.handle(VerboseEvent({"message": f"STDOUT:\n{proc.stdout.strip()}"}))
            else:
                event_handler.handle(VerboseEvent({"message": "No STDOUT"}))
            if proc.stderr.strip() != "":
                event_handler.handle(VerboseEvent({"message": f"STDERR:\n{proc.stderr.strip()}"}))
            else:
                event_handler.handle(VerboseEvent({"message": "No STDERR"}))
            proc.check_returncode()



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""Commands take a Batch that has already been transformed and perform post-processing
steps before the Batch is submitted, such as code generation. These steps can be performed
either pre or post-validation, depending on if the run_pre_validation flag is set."""



# AutoTransform
# Large scale, component based code modification library
#
# Licensed under the MIT License <http://opensource.org/licenses/MIT>
# SPDX-License-Identifier: MIT
# Copyright (c) 2022-present Nathan Rockenbach <http://github.com/nathro>

# @black_format

"""The base class and associated classes for Input components."""

from __future__ import annotations

from abc import abstractmethod
from enum import Enum
from typing import ClassVar, Sequence

from autotransform.item.base import Item
from autotransform.util.component import ComponentFactory, ComponentImport, NamedComponent


class InputName(str, Enum):
    """A simple enum for mapping."""

    DIRECTORY = "directory"
    EMPTY = "empty"
    GIT_GREP = "git_grep"
    INLINE = "inline"
    INLINE_FILE = "inline_file"
    INLINE_GENERIC = "inline_generic"
    SCRIPT = "script"


class Input(NamedComponent):
    """The base for Input components. Used by AutoTransform to get Items that
    represent potentially transformable units for a Schema. Usually returns files but
    any Item can be returned as long as Schema components work with it.

    Attributes:
        name (ClassVar[InputName]): The name of the component.
    """

    name: ClassVar[InputName]

    @abstractmethod
    def get_items(self) -> Sequence[Item]:
        """Get a list of Items to be used by the transformation based on the Input criteria. Usually
        returns FileItems.

        Returns:
            Sequence[Item]: The eligible Items for transformation.
        """


FACTORY = ComponentFactory(
    {
        InputName.DIRECTORY: ComponentImport(
            class_name="DirectoryInput", module="autotransform.input.directory"
        ),
        InputName.EMPTY: ComponentImport(
            class_name="EmptyInput", module="autotransform.input.empty"
        ),
        InputName.GIT_GREP: ComponentImport(
            class_name="GitGrepInput", module="autotransform.input.gitgrep"
        ),
        InputName.INLINE: ComponentImport(
            class_name="InlineInput", module="autotransform.input.inline"
        ),
        InputName.INLINE_FILE: ComponentImport(
            class_name="InlineFileInput", module="autotransform.input.inline"
        ),
        InputName.INLINE_GENERIC: ComponentImport(
            class_name="InlineGenericInput", module="autotransform.input.inline"
        ),
        InputName.SCRIPT: ComponentImport(
            class_name="ScriptInput", module="autotransform.input.script"
        ),
    },
    Input,  # type: ignore [type-abstract]
    "input.json",
)
